@@ -1,9 +1,9 @@
 /*  Lunzip - Decompressor for lzip files
-    Copyright (C) 2010, 2011, 2012, 2013 Antonio Diaz Diaz.
+    Copyright (C) 2010, 2011, 2012, 2013, 2014 Antonio Diaz Diaz.
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
-    the Free Software Foundation, either version 3 of the License, or
+    the Free Software Foundation, either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
@@ -45,7 +45,7 @@ void Pp_show_msg( struct Pretty_print * const pp, const char * const msg )
       for( i = 0; i < len; ++i ) fprintf( stderr, " " );
       if( !msg ) fflush( stderr );
       }
-    if( msg ) fprintf( stderr, "%s.\n", msg );
+    if( msg ) fprintf( stderr, "%s\n", msg );
     }
   }
 
@@ -55,17 +55,17 @@ void Pp_show_msg( struct Pretty_print * const pp, const char * const msg )
 */
 static int readblock( const int fd, uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = read( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
+    const int n = read( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
     else if( n == 0 ) break;				/* EOF */
-    else if( errno != EINTR && errno != EAGAIN ) break;
+    else if( errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
   }
 
 
@@ -74,16 +74,25 @@ static int readblock( const int fd, uint8_t * const buf, const int size )
 */
 static int writeblock( const int fd, const uint8_t * const buf, const int size )
   {
-  int rest = size;
+  int sz = 0;
   errno = 0;
-  while( rest > 0 )
+  while( sz < size )
     {
-    const int n = write( fd, buf + size - rest, rest );
-    if( n > 0 ) rest -= n;
-    else if( n < 0 && errno != EINTR && errno != EAGAIN ) break;
+    const int n = write( fd, buf + sz, size - sz );
+    if( n > 0 ) sz += n;
+    else if( n < 0 && errno != EINTR ) break;
     errno = 0;
     }
-  return size - rest;
+  return sz;
+  }
+
+
+int seek_read( const int fd, uint8_t * const buf, const int size,
+               const int offset )
+  {
+  if( lseek( fd, offset, SEEK_END ) >= 0 )
+    return readblock( fd, buf, size );
+  return 0;
   }
 
 
@@ -102,31 +111,33 @@ bool Rd_read_block( struct Range_decoder * const rdec )
   }
 
 
-void LZd_flush_data( struct LZ_decoder * const decoder )
+void LZd_flush_data( struct LZ_decoder * const d )
   {
-  if( decoder->pos > decoder->stream_pos )
+  if( d->pos > d->stream_pos )
     {
-    const int size = decoder->pos - decoder->stream_pos;
-    CRC32_update_buf( &decoder->crc, decoder->buffer + decoder->stream_pos, size );
-    if( decoder->outfd >= 0 &&
-        writeblock( decoder->outfd, decoder->buffer + decoder->stream_pos, size ) != size )
+    const int size = d->pos - d->stream_pos;
+    CRC32_update_buf( &d->crc, d->buffer + d->stream_pos, size );
+    if( d->outfd >= 0 &&
+        writeblock( d->outfd, d->buffer + d->stream_pos, size ) != size )
       { show_error( "Write error", errno, false ); cleanup_and_fail( 1 ); }
-    if( decoder->pos >= decoder->buffer_size )
-      { decoder->partial_data_pos += decoder->pos; decoder->pos = 0; }
-    decoder->stream_pos = decoder->pos;
+    if( d->pos >= d->buffer_size )
+      { d->partial_data_pos += d->pos; d->pos = 0; }
+    d->stream_pos = d->pos;
     }
   }
 
 
-bool LZd_verify_trailer( struct LZ_decoder * const decoder,
-                         struct Pretty_print * const pp )
+static bool LZd_verify_trailer( struct LZ_decoder * const d,
+                                struct Pretty_print * const pp )
   {
   File_trailer trailer;
-  const unsigned long long member_size =
-    Rd_member_position( decoder->rdec ) + Ft_size;
+  const unsigned long long member_size = Rd_member_position( d->rdec ) + Ft_size;
+  unsigned long long trailer_data_size;
+  unsigned long long trailer_member_size;
+  unsigned trailer_crc;
   bool error = false;
 
-  int size = Rd_read_data( decoder->rdec, trailer, Ft_size );
+  int size = Rd_read_data( d->rdec, trailer, Ft_size );
   if( size < Ft_size )
     {
     error = true;
@@ -139,60 +150,66 @@ bool LZd_verify_trailer( struct LZ_decoder * const decoder,
     while( size < Ft_size ) trailer[size++] = 0;
     }
 
-  if( decoder->rdec->code != 0 )
+  if( d->rdec->code != 0 )
     {
     error = true;
-    Pp_show_msg( pp, "Range decoder final code is not zero" );
+    Pp_show_msg( pp, "Range decoder final code is not zero." );
     }
-  if( Ft_get_data_crc( trailer ) != LZd_crc( decoder ) )
+  trailer_crc = Ft_get_data_crc( trailer );
+  if( trailer_crc != LZd_crc( d ) )
     {
     error = true;
     if( verbosity >= 0 )
       {
       Pp_show_msg( pp, 0 );
       fprintf( stderr, "CRC mismatch; trailer says %08X, data CRC is %08X.\n",
-               Ft_get_data_crc( trailer ), LZd_crc( decoder ) );
+               trailer_crc, LZd_crc( d ) );
       }
     }
-  if( Ft_get_data_size( trailer ) != LZd_data_position( decoder ) )
+  trailer_data_size = Ft_get_data_size( trailer );
+  if( trailer_data_size != LZd_data_position( d ) )
     {
     error = true;
     if( verbosity >= 0 )
       {
       Pp_show_msg( pp, 0 );
       fprintf( stderr, "Data size mismatch; trailer says %llu, data size is %llu (0x%llX).\n",
-               Ft_get_data_size( trailer ), LZd_data_position( decoder ), LZd_data_position( decoder ) );
+               trailer_data_size, LZd_data_position( d ), LZd_data_position( d ) );
       }
     }
-  if( Ft_get_member_size( trailer ) != member_size )
+  trailer_member_size = Ft_get_member_size( trailer );
+  if( trailer_member_size != member_size )
     {
     error = true;
     if( verbosity >= 0 )
       {
       Pp_show_msg( pp, 0 );
       fprintf( stderr, "Member size mismatch; trailer says %llu, member size is %llu (0x%llX).\n",
-               Ft_get_member_size( trailer ), member_size, member_size );
+               trailer_member_size, member_size, member_size );
       }
     }
-  if( !error && verbosity >= 2 && LZd_data_position( decoder ) > 0 && member_size > 0 )
+  if( !error && verbosity >= 2 && LZd_data_position( d ) > 0 && member_size > 0 )
     fprintf( stderr, "%6.3f:1, %6.3f bits/byte, %5.2f%% saved.  ",
-             (double)LZd_data_position( decoder ) / member_size,
-             ( 8.0 * member_size ) / LZd_data_position( decoder ),
-             100.0 * ( 1.0 - ( (double)member_size / LZd_data_position( decoder ) ) ) );
+             (double)LZd_data_position( d ) / member_size,
+             ( 8.0 * member_size ) / LZd_data_position( d ),
+             100.0 * ( 1.0 - ( (double)member_size / LZd_data_position( d ) ) ) );
   if( !error && verbosity >= 4 )
     fprintf( stderr, "data CRC %08X, data size %9llu, member size %8llu.  ",
-             Ft_get_data_crc( trailer ),
-             Ft_get_data_size( trailer ), Ft_get_member_size( trailer ) );
+             trailer_crc, trailer_data_size, trailer_member_size );
   return !error;
   }
 
 
 /* Return value: 0 = OK, 1 = decoder error, 2 = unexpected EOF,
                  3 = trailer error, 4 = unknown marker found. */
-int LZd_decode_member( struct LZ_decoder * const decoder,
+int LZd_decode_member( struct LZ_decoder * const d,
                        struct Pretty_print * const pp )
   {
-  struct Range_decoder * const rdec = decoder->rdec;
+  struct Range_decoder * const rdec = d->rdec;
+  void (* const copy_block)
+       ( struct LZ_decoder * const d, const int distance, int len ) =
+    ( (unsigned)d->buffer_size >= d->dictionary_size ) ?
+    &LZd_copy_block : &LZd_copy_block2;
   unsigned rep0 = 0;		/* rep[0-3] latest four distances */
   unsigned rep1 = 0;		/* used for efficient coding of */
   unsigned rep2 = 0;		/* repeated distances */
@@ -202,37 +219,37 @@ int LZd_decode_member( struct LZ_decoder * const decoder,
   Rd_load( rdec );
   while( !Rd_finished( rdec ) )
     {
-    const int pos_state = LZd_data_position( decoder ) & pos_state_mask;
-    if( Rd_decode_bit( rdec, &decoder->bm_match[state][pos_state] ) == 0 )	/* 1st bit */
+    const int pos_state = LZd_data_position( d ) & pos_state_mask;
+    if( Rd_decode_bit( rdec, &d->bm_match[state][pos_state] ) == 0 )	/* 1st bit */
       {
-      const uint8_t prev_byte = LZd_get_prev_byte( decoder );
+      const uint8_t prev_byte = LZd_get_prev_byte( d );
       if( St_is_char( state ) )
         {
         state -= ( state < 4 ) ? state : 3;
-        LZd_put_byte( decoder, Rd_decode_tree( rdec,
-                      decoder->bm_literal[get_lit_state(prev_byte)], 8 ) );
+        LZd_put_byte( d, Rd_decode_tree( rdec,
+                         d->bm_literal[get_lit_state(prev_byte)], 8 ) );
         }
       else
         {
         state -= ( state < 10 ) ? 3 : 6;
-        LZd_put_byte( decoder, Rd_decode_matched( rdec,
-                      decoder->bm_literal[get_lit_state(prev_byte)],
-                      LZd_get_byte( decoder, rep0 ) ) );
+        LZd_put_byte( d, Rd_decode_matched( rdec,
+                         d->bm_literal[get_lit_state(prev_byte)],
+                         LZd_get_byte( d, rep0 ) ) );
         }
       }
     else
       {
       int len;
-      if( Rd_decode_bit( rdec, &decoder->bm_rep[state] ) != 0 )	/* 2nd bit */
+      if( Rd_decode_bit( rdec, &d->bm_rep[state] ) != 0 )	/* 2nd bit */
         {
-        if( Rd_decode_bit( rdec, &decoder->bm_rep0[state] ) != 0 )	/* 3rd bit */
+        if( Rd_decode_bit( rdec, &d->bm_rep0[state] ) != 0 )	/* 3rd bit */
           {
           unsigned distance;
-          if( Rd_decode_bit( rdec, &decoder->bm_rep1[state] ) == 0 )	/* 4th bit */
+          if( Rd_decode_bit( rdec, &d->bm_rep1[state] ) == 0 )	/* 4th bit */
             distance = rep1;
           else
             {
-            if( Rd_decode_bit( rdec, &decoder->bm_rep2[state] ) == 0 )	/* 5th bit */
+            if( Rd_decode_bit( rdec, &d->bm_rep2[state] ) == 0 )	/* 5th bit */
               distance = rep2;
             else
               { distance = rep3; rep3 = rep2; }
@@ -243,19 +260,19 @@ int LZd_decode_member( struct LZ_decoder * const decoder,
           }
         else
           {
-          if( Rd_decode_bit( rdec, &decoder->bm_len[state][pos_state] ) == 0 )	/* 4th bit */
+          if( Rd_decode_bit( rdec, &d->bm_len[state][pos_state] ) == 0 )	/* 4th bit */
             { state = St_set_short_rep( state );
-              LZd_put_byte( decoder, LZd_get_byte( decoder, rep0 ) ); continue; }
+              LZd_put_byte( d, LZd_get_byte( d, rep0 ) ); continue; }
           }
         state = St_set_rep( state );
-        len = min_match_len + Rd_decode_len( rdec, &decoder->rep_len_model, pos_state );
+        len = min_match_len + Rd_decode_len( rdec, &d->rep_len_model, pos_state );
         }
       else
         {
         int dis_slot;
         const unsigned rep0_saved = rep0;
-        len = min_match_len + Rd_decode_len( rdec, &decoder->match_len_model, pos_state );
-        dis_slot = Rd_decode_tree6( rdec, decoder->bm_dis_slot[get_len_state(len)] );
+        len = min_match_len + Rd_decode_len( rdec, &d->match_len_model, pos_state );
+        dis_slot = Rd_decode_tree6( rdec, d->bm_dis_slot[get_len_state(len)] );
         if( dis_slot < start_dis_model ) rep0 = dis_slot;
         else
           {
@@ -263,19 +280,19 @@ int LZd_decode_member( struct LZ_decoder * const decoder,
           rep0 = ( 2 | ( dis_slot & 1 ) ) << direct_bits;
           if( dis_slot < end_dis_model )
             rep0 += Rd_decode_tree_reversed( rdec,
-                    decoder->bm_dis + rep0 - dis_slot - 1, direct_bits );
+                    d->bm_dis + rep0 - dis_slot - 1, direct_bits );
           else
             {
             rep0 += Rd_decode( rdec, direct_bits - dis_align_bits ) << dis_align_bits;
-            rep0 += Rd_decode_tree_reversed4( rdec, decoder->bm_align );
+            rep0 += Rd_decode_tree_reversed4( rdec, d->bm_align );
             if( rep0 == 0xFFFFFFFFU )		/* Marker found */
               {
               rep0 = rep0_saved;
               Rd_normalize( rdec );
-              LZd_flush_data( decoder );
+              LZd_flush_data( d );
               if( len == min_match_len )	/* End Of Stream marker */
                 {
-                if( LZd_verify_trailer( decoder, pp ) ) return 0; else return 3;
+                if( LZd_verify_trailer( d, pp ) ) return 0; else return 3;
                 }
               if( len == min_match_len + 1 )	/* Sync Flush marker */
                 {
@@ -292,13 +309,12 @@ int LZd_decode_member( struct LZ_decoder * const decoder,
           }
         rep3 = rep2; rep2 = rep1; rep1 = rep0_saved;
         state = St_set_match( state );
-        if( rep0 >= (unsigned)decoder->dictionary_size ||
-            ( rep0 >= (unsigned)decoder->pos && !decoder->partial_data_pos ) )
-          { LZd_flush_data( decoder ); return 1; }
+        if( rep0 >= d->dictionary_size || rep0 >= LZd_data_position( d ) )
+          { LZd_flush_data( d ); return 1; }
         }
-      LZd_copy_block( decoder, rep0, len );
+      copy_block( d, rep0, len );
       }
     }
-  LZd_flush_data( decoder );
+  LZd_flush_data( d );
   return 2;
   }
